import pytest
from gwcloud_python.utils import file_filters
from gwcloud_python import FileReference, FileReferenceList


@pytest.fixture
def png_data():
    return FileReferenceList([
        FileReference(path='data/dir/test1.png', file_size='1', download_token='test_download_token_1'),
        FileReference(path='data/dir/test2.png', file_size='1', download_token='test_download_token_2'),
    ])


@pytest.fixture
def png_result():
    return FileReferenceList([
        FileReference(path='result/dir/test1.png', file_size='1', download_token='test_download_token_3'),
        FileReference(path='result/dir/test2.png', file_size='1', download_token='test_download_token_4'),
    ])


@pytest.fixture
def png_extra_no_dir():
    return FileReferenceList([
        FileReference(path='test1.png', file_size='1', download_token='test_download_token_5'),
        FileReference(path='test2.png', file_size='1', download_token='test_download_token_6'),
    ])


@pytest.fixture
def png_extra_dir():
    return FileReferenceList([
        FileReference(path='arbitrary/dir/test1.png', file_size='1', download_token='test_download_token_7'),
        FileReference(path='arbitrary/dir/test2.png', file_size='1', download_token='test_download_token_8'),
    ])


@pytest.fixture
def corner():
    return FileReferenceList([
        FileReference(path='test1_corner.png', file_size='1', download_token='test_download_token_9'),
        FileReference(path='test2_corner.png', file_size='1', download_token='test_download_token_10'),
    ])


@pytest.fixture
def index():
    return FileReferenceList([
        FileReference(path='index.html', file_size='1', download_token='test_download_token_11'),
    ])


@pytest.fixture
def config():
    return FileReferenceList([
        FileReference(path='a_config_complete.ini', file_size='1', download_token='test_download_token_12'),
    ])


@pytest.fixture
def merge():
    return FileReferenceList([
        FileReference(path='result/dir/a_merge_result.json', file_size='1', download_token='test_download_token_13'),
    ])


@pytest.fixture
def unmerge():
    return FileReferenceList([
        FileReference(path='result/dir/a_result.json', file_size='1', download_token='test_download_token_14'),
    ])


@pytest.fixture
def png(png_data, png_result, png_extra_no_dir, png_extra_dir, corner):
    return png_data + png_result + png_extra_no_dir + png_extra_dir + corner


@pytest.fixture
def default_with_merge(png_data, png_result, index, config, merge):
    return png_data + png_result + index + config + merge


@pytest.fixture
def full_with_merge(png, index, config, merge, unmerge):
    return png + index + config + merge + unmerge


@pytest.fixture
def default_without_merge(png_data, png_result, index, config, unmerge):
    return png_data + png_result + index + config + unmerge


@pytest.fixture
def full_without_merge(png, index, config, unmerge):
    return png + index + config + unmerge


def test_default_file_filter(full_with_merge, default_with_merge, full_without_merge, default_without_merge):
    sub_list = file_filters.default_filter(full_with_merge)
    assert file_filters.sort_file_list(sub_list) == file_filters.sort_file_list(default_with_merge)

    sub_list = file_filters.default_filter(full_without_merge)
    assert file_filters.sort_file_list(sub_list) == file_filters.sort_file_list(default_without_merge)


def test_png_file_filter(full_with_merge, png):
    sub_list = file_filters.png_filter(full_with_merge)
    assert file_filters.sort_file_list(sub_list) == file_filters.sort_file_list(png)


def test_config_file_filter(full_with_merge, config):
    sub_list = file_filters.config_filter(full_with_merge)
    assert file_filters.sort_file_list(sub_list) == file_filters.sort_file_list(config)


def test_corner_file_filter(full_with_merge, corner):
    sub_list = file_filters.corner_plot_filter(full_with_merge)
    assert file_filters.sort_file_list(sub_list) == file_filters.sort_file_list(corner)


def test_custom_path_file_filter(full_without_merge, merge, png_data, png_extra_dir, index):
    # Test here is not exactly comprehensive, but it's good enough for now
    full = full_without_merge + merge
    sub_list = file_filters.custom_path_filter(full, directory='data')
    assert file_filters.sort_file_list(sub_list) == file_filters.sort_file_list(png_data)
    sub_list = file_filters.custom_path_filter(full, name='merge')
    assert file_filters.sort_file_list(sub_list) == file_filters.sort_file_list(merge)
    sub_list = file_filters.custom_path_filter(full, extension='html')
    assert file_filters.sort_file_list(sub_list) == file_filters.sort_file_list(index)

    sub_list = file_filters.custom_path_filter(full, directory='result', name='merge')
    assert file_filters.sort_file_list(sub_list) == file_filters.sort_file_list(merge)
    sub_list = file_filters.custom_path_filter(full, directory='arbitrary', extension='png')
    assert file_filters.sort_file_list(sub_list) == file_filters.sort_file_list(png_extra_dir)
    sub_list = file_filters.custom_path_filter(full, name='merge', extension='json')
    assert file_filters.sort_file_list(sub_list) == file_filters.sort_file_list(merge)

    sub_list = file_filters.custom_path_filter(full, directory='result', name='merge', extension='json')
    assert file_filters.sort_file_list(sub_list) == file_filters.sort_file_list(merge)



from . import identifiers
from functools import partial


def _split_file_list(identifier, file_list):
    matches = []
    others = []
    for f in file_list:
        if identifier(f.path):
            matches.append(f)
        else:
            others.append(f)
    return matches, others


def _filter_file_list(identifier, file_list):
    return [f for f in file_list if identifier(f.path)]


def _match_all(identifier_list, file_list):
    output = file_list
    for identifier in identifier_list:
        output, _ = _split_file_list(identifier, output)

    return output


def default_filter(file_list):
    """Takes an input file list and returns a subset of that file list containing:

    - Any HTML file
    - Any file ending with '_config_complete.ini'
    - Any PNG files in the 'data' directory
    - Any PNG files in the 'result' directory
    - Any file in the 'result' directory ending in '_merge_result.json', or '_result.json' if there is no merged file

    Parameters
    ----------
    file_list : .FileReferenceList
        A list of FileReference objects which will be filtered

    Returns
    -------
    .FileReferenceList
        Subset of the input FileReferenceList containing only the paths that match the above default file criteria
    """
    # Get png files in data dir
    data_png_file_list, file_list = _split_file_list(identifiers.data_png_file, file_list)

    # Get png files in result dir
    result_png_file_list, file_list = _split_file_list(identifiers.result_png_file, file_list)

    # Get complete config file
    config_file_list, file_list = _split_file_list(identifiers.config_file, file_list)

    # Get index html file
    html_file_list, file_list = _split_file_list(identifiers.html_file, file_list)

    # Get merged json file in result dir
    result_json_file_list, file_list = _split_file_list(identifiers.result_merged_json_file, file_list)

    # If merged json doesn't exist, get result json file in result dir
    if not result_json_file_list:
        result_json_file_list, file_list = _split_file_list(identifiers.result_json_file, file_list)

    return data_png_file_list + result_png_file_list + config_file_list + html_file_list + result_json_file_list


def config_filter(file_list):
    """Takes an input file list and returns a subset of that file list containing:

    - Any file ending with '_config_complete.ini'

    Parameters
    ----------
    file_list : .FileReferenceList
        A list of FileReference objects which will be filtered

    Returns
    -------
    .FileReferenceList
        Subset of the input FileReferenceList containing only the paths that match the above config file criteria
    """
    return _filter_file_list(identifiers.config_file, file_list)


def png_filter(file_list):
    """Takes an input file list and returns a subset of that file list containing:

    - Any PNG file

    Parameters
    ----------
    file_list : .FileReferenceList
        A list of FileReference objects which will be filtered

    Returns
    -------
    .FileReferenceList
        Subset of the input FileReferenceList containing only the paths that match the above png file criteria
    """
    return _filter_file_list(identifiers.png_file, file_list)


def corner_plot_filter(file_list):
    """Takes an input file list and returns a subset of that file list containing:

    - Any file ending in '_corner.png'

    Parameters
    ----------
    file_list : .FileReferenceList
        A list of FileReference objects which will be filtered

    Returns
    -------
    .FileReferenceList
        Subset of the input FileReferenceList containing only the paths that match the above corner plot file criteria
    """
    return _filter_file_list(identifiers.corner_plot_file, file_list)


def custom_path_filter(file_list, directory=None, name=None, extension=None):
    """Takes an input file list and returns a subset of that file list containing:

    - Any file that has any enclosing directory matching the `directory` argument
    - Any file that has any part of its filename matching the `name` argument
    - Any file that has an extension matching the `extension` argument

    Parameters
    ----------
    file_list : .FileReferenceList
        A list of FileReference objects which will be filtered
    directory : str, optional
        Directory to match, by default None
    name : str, optional
        Part of filename to match, by default None
    extension : str, optional
        File extension to match, by default None

    Returns
    -------
    .FileReferenceList
        Subset of the input FileReferenceList containing only the paths that match the above corner plot file criteria
    """
    identifier_list = []
    if directory:
        identifier_list.append(partial(identifiers._file_dir, directory=str(directory)))
    if name:
        identifier_list.append(partial(identifiers._file_name, name=str(name)))
    if extension:
        identifier_list.append(partial(identifiers._file_suffix, suffix=str(extension)))

    return _match_all(identifier_list, file_list)


def sort_file_list(file_list):
    """Sorts a file list based on the 'path' key of the dicts. Primarily used for equality checks.

    Parameters
    ----------
    file_list : .FileReferenceList
        A list of FileReference objects which will be filtered

    Returns
    -------
    .FileReferenceList
        A FileReferenceList containing the same members as the input,
        sorted by the path attribute of the FileReference objects
    """
    return sorted(file_list, key=lambda f: f.path)



import concurrent.futures
import logging
import tarfile
from functools import partial
from pathlib import Path
from tempfile import NamedTemporaryFile

import requests
from gwdc_python import GWDC
from tqdm import tqdm

from .bilby_job import BilbyJob
from .event_id import EventID
from .file_reference import FileReference, FileReferenceList
from .helpers import TimeRange, Cluster
from .utils import convert_dict_keys

GWCLOUD_ENDPOINT = 'https://gwcloud.org.au/bilby/graphql'
GWCLOUD_FILE_DOWNLOAD_ENDPOINT = 'https://gwcloud.org.au/job/apiv1/file/?fileId='
GWCLOUD_UPLOADED_JOB_FILE_DOWNLOAD_ENDPOINT = 'https://gwcloud.org.au/bilby/file_download/?fileId='

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
logger.addHandler(ch)


class GWCloud:
    """
    GWCloud class provides an API for interacting with Bilby, allowing jobs to be submitted and acquired.

    Parameters
    ----------
    token : str
        API token for a Bilby user
    endpoint : str, optional
        URL to which we send the queries, by default GWCLOUD_ENDPOINT

    Attributes
    ----------
    client : GWDC
        Handles a lot of the underlying logic surrounding the queries
    """
    def __init__(self, token, endpoint=GWCLOUD_ENDPOINT):
        self.client = GWDC(token=token, endpoint=endpoint)
        self.request = self.client.request  # Setting shorthand for simplicity

    def start_bilby_job_from_string(self, job_name, job_description, private, ini_string, cluster=Cluster.DEFAULT):
        """Submit the parameters required to start a Bilby job, using the contents of an .ini file

        Parameters
        ----------
        job_name : str
            Name of the job to be created
        job_description : str
            Description of the job to be created
        private : bool
            True if job should be private, False if public
        ini_string : str
            The contents of a Bilby ini file
        cluster : .Cluster or str
            The name of the cluster to submit the job to

        Returns
        -------
        str
            Message received from server after job submission
        """
        query = """
            mutation NewBilbyJobFromIniString($input: BilbyJobFromIniStringMutationInput!){
                newBilbyJobFromIniString (input: $input) {
                    result {
                        jobId
                    }
                }
            }
        """

        variables = {
            "input": {
                "params": {
                    "details": {
                        "name": job_name,
                        "description": job_description,
                        "private": private,
                        "cluster": cluster.value if isinstance(cluster, Cluster) else cluster
                    },
                    "iniString": {
                        "iniString": str(ini_string)
                    }
                },
            }
        }

        data = self.request(query=query, variables=variables)
        job_id = data['newBilbyJobFromIniString']['result']['jobId']
        return self.get_job_by_id(job_id)

    def start_bilby_job_from_file(self, job_name, job_description, private, ini_file, cluster=Cluster.DEFAULT):
        """Submit the parameters required to start a Bilby job, using an .ini file

        Parameters
        ----------
        job_name : str
            Name of the job to be created
        job_description : str
            Description of the job to be created
        private : bool
            True if job should be private, False if public
        ini_file : str or Path
            Path to an .ini file for running a Bilby job
        cluster : .Cluster or str
            The name of the cluster to submit the job to

        Returns
        -------
        str
            Message received from server after job submission
        """
        ini_file = Path(ini_file)
        with ini_file.open() as f:
            ini_string = f.read().strip()
            return self.start_bilby_job_from_string(job_name, job_description, private, ini_string, cluster)

    def get_preferred_job_list(self, search=""):
        """Get list of public Bilby jobs corresponding to a search of "preferred" and a time_range of "Any time"

        Parameters
        ----------
        search : str, optional
            Search terms by which to filter public job list, by default ""

        Returns
        -------
        list
            List of BilbyJob instances for the preferred jobs corresponding to the search terms
        """
        return self.get_public_job_list(search=f"preferred lasky {search}", time_range=TimeRange.ANY)

    def _get_job_model_from_query(self, query_data):
        if not query_data:
            # logger.info('No')
            return None
        return BilbyJob(
            client=self,
            **convert_dict_keys(
                query_data,
                [
                    ('id', 'job_id'),
                ]
            )
        )

    def get_public_job_list(self, search="", time_range=TimeRange.ANY, number=100):
        """Obtains a list of public Bilby jobs, filtering based on the search terms
        and the time range within which the job was created.

        Parameters
        ----------
        search : str, optional
            Search terms by which to filter public job list, by default ""
        time_range : .TimeRange or str, optional
            Time range by which to filter job list, by default TimeRange.ANY
        number : int, optional
            Number of job results to return in one request, by default 100

        Returns
        -------
        list
            List of BilbyJob instances for the jobs corresponding to the search terms and in the specified time range
        """
        query = """
            query ($search: String, $timeRange: String, $first: Int){
                publicBilbyJobs (search: $search, timeRange: $timeRange, first: $first) {
                    edges {
                        node {
                            id
                            user
                            name
                            description
                            jobStatus {
                                name
                                date
                            }
                        }
                    }
                }
            }
        """

        variables = {
            "search": search,
            "timeRange": time_range.value if isinstance(time_range, TimeRange) else time_range,
            "first": number
        }

        data = self.request(query=query, variables=variables)

        if not data['publicBilbyJobs']['edges']:
            logger.info('Job search returned no results.')
            return []

        return [self._get_job_model_from_query(job['node']) for job in data['publicBilbyJobs']['edges']]

    def get_job_by_id(self, job_id):
        """Get a Bilby job instance corresponding to a specific job ID

        Parameters
        ----------
        job_id : str
            ID of job to obtain

        Returns
        -------
        BilbyJob
            BilbyJob instance corresponding to the input ID
        """
        query = """
            query ($id: ID!){
                bilbyJob (id: $id) {
                    id
                    name
                    userId
                    description
                    jobStatus {
                        name
                        date
                    }
                    eventId {
                        eventId
                        triggerId
                        nickname
                        isLigoEvent
                    }
                }
            }
        """

        variables = {
            "id": job_id
        }

        data = self.request(query=query, variables=variables)

        if not data['bilbyJob']:
            logger.info('No job matching input ID was returned.')
            return None

        return self._get_job_model_from_query(data['bilbyJob'])

    def get_user_jobs(self, number=100):
        """Obtains a list of Bilby jobs created by the user, filtering based on the search terms
        and the time range within which the job was created.

        Parameters
        ----------
        number : int, optional
            Number of job results to return in one request, by default 100

        Returns
        -------
        list
            List of BilbyJob instances for the jobs corresponding to the search terms and in the specified time range
        """
        query = """
            query ($first: Int){
                bilbyJobs (first: $first){
                    edges {
                        node {
                            id
                            name
                            userId
                            description
                            jobStatus {
                                name
                                date
                            }
                            eventId {
                                eventId
                                triggerId
                                nickname
                                isLigoEvent
                            }
                        }
                    }
                }
            }
        """

        variables = {
            "first": number
        }

        data = self.request(query=query, variables=variables)

        return [self._get_job_model_from_query(job['node']) for job in data['bilbyJobs']['edges']]

    def _get_files_by_job_id(self, job_id):
        query = """
            query ($jobId: ID!) {
                bilbyResultFiles (jobId: $jobId) {
                    files {
                        path
                        isDir
                        fileSize
                        downloadToken
                    }
                    isUploadedJob
                }
            }
        """

        variables = {
            "jobId": job_id
        }

        data = self.request(query=query, variables=variables)

        file_list = FileReferenceList()
        for f in data['bilbyResultFiles']['files']:
            if f['isDir']:
                continue
            f.pop('isDir')
            file_list.append(FileReference(**convert_dict_keys(f)))

        return file_list, data['bilbyResultFiles']['isUploadedJob']

    def _get_file_map_fn(self, file_id, file_path, progress_bar, is_uploaded_job=False):
        endpoint = GWCLOUD_FILE_DOWNLOAD_ENDPOINT \
            if not is_uploaded_job else \
            GWCLOUD_UPLOADED_JOB_FILE_DOWNLOAD_ENDPOINT

        download_url = endpoint + str(file_id)
        content = b''
        with requests.get(download_url, stream=True) as request:
            for chunk in request.iter_content(chunk_size=1024 * 16, decode_unicode=True):
                progress_bar.update(len(chunk))
                content += chunk
        return (file_path, content)

    def _get_files_by_reference(self, job_id, file_references, is_uploaded_job=False):
        """Obtains file data when provided a job ID and a FileReferenceList

        Parameters
        ----------
        job_id : str
            Job ID
        file_references : FileReferenceList
            Contains the FileReferences objects for which to download the contents

        Returns
        -------
        list
            List of tuples containing the file path and file contents as a byte string
        """
        file_ids = self._get_download_ids_from_tokens(job_id, file_references.get_tokens())
        file_paths = file_references.get_paths()

        with concurrent.futures.ThreadPoolExecutor(max_workers=20) as executor:
            progress = tqdm(total=file_references.get_total_bytes(), leave=True, unit='B', unit_scale=True)
            files = list(
                executor.map(
                    partial(
                        self._get_file_map_fn,
                        progress_bar=progress,
                        is_uploaded_job=is_uploaded_job
                    ), file_ids, file_paths
                )
            )
            progress.close()
            logger.info(f'All {len(file_ids)} files downloaded!')

        return files

    def _save_file_map_fn(self, file_id, file_path, progress_bar, is_uploaded_job=False):
        endpoint = GWCLOUD_FILE_DOWNLOAD_ENDPOINT \
            if not is_uploaded_job else \
            GWCLOUD_UPLOADED_JOB_FILE_DOWNLOAD_ENDPOINT

        download_url = endpoint + str(file_id)
        file_path.parents[0].mkdir(parents=True, exist_ok=True)

        with requests.get(download_url, stream=True) as request:
            with file_path.open("wb+") as f:
                for chunk in request.iter_content(chunk_size=1024 * 16):
                    progress_bar.update(len(chunk))
                    f.write(chunk)

    def _save_files_by_reference(self, job_id, file_references, root_path, preserve_directory_structure=True,
                                 is_uploaded_job=False):
        """Save files when provided a job ID and a FileReferenceList

        Parameters
        ----------
        job_id : str
            Job ID
        file_references : FileReferenceList
            Contains the FileReference objects for which to save the associated files
        root_path : str or ~pathlib.Path
            Directory into which to save the files
        preserve_directory_structure : bool, optional
            Remove any directory structure for the downloaded files, by default True

        Returns
        -------
        str
            Success message
        """
        file_ids = self._get_download_ids_from_tokens(job_id, file_references.get_tokens())
        file_paths = file_references.get_output_paths(root_path, preserve_directory_structure)

        with concurrent.futures.ThreadPoolExecutor(max_workers=20) as executor:
            progress = tqdm(total=file_references.get_total_bytes(), leave=True, unit='B', unit_scale=True)
            list(
                executor.map(
                    partial(
                        self._save_file_map_fn,
                        progress_bar=progress,
                        is_uploaded_job=is_uploaded_job
                    ),
                    file_ids, file_paths
                )
            )
            progress.close()
            logger.info(f'All {len(file_ids)} files saved!')

        return 'Success'

    def _get_download_id_from_token(self, job_id, file_token):
        """Get a single file download id for a file download token

        Parameters
        ----------
        job_id : str
            Job id which owns the file token

        file_token : str
            Download token for the desired file

        Returns
        -------
        str
            Download id for the desired file
        """
        return self._get_download_ids_from_tokens(job_id, [file_token])[0]

    def _get_download_ids_from_tokens(self, job_id, file_tokens):
        """Get many file download ids for a list of file download tokens

        Parameters
        ----------
        job_id : str
            Job id which owns the file token

        file_tokens : list
            Download tokens for the desired files

        Returns
        -------
        list
            List of download ids for the desired files
        """
        query = """
            mutation ResultFileMutation($input: GenerateFileDownloadIdsInput!) {
                generateFileDownloadIds(input: $input) {
                    result
                }
            }
        """

        variables = {
            "input": {
                "jobId": job_id,
                "downloadTokens": file_tokens
            }
        }

        data = self.request(query=query, variables=variables)

        return data['generateFileDownloadIds']['result']

    def _generate_upload_token(self):
        """Creates a new long lived upload token for use uploading jobs

        Returns
        -------
        str
            The upload token
        """
        query = """
            query GenerateBilbyJobUploadToken {
                generateBilbyJobUploadToken {
                  token
                }
            }
        """

        data = self.request(query=query)
        return data['generateBilbyJobUploadToken']['token']

    def upload_job_archive(self, description, job_archive, public=False):
        """Upload a bilby job to GWCloud by job output archive

        Parameters
        ----------
        description : str
            The description of the job to add to the database

        public : bool
            If the uploaded job should be public or not

        job_archive : str
            The path to the job output archive to upload

        Returns
        -------
        BilbyJob
            The created Bilby job
        """
        query = """
            mutation JobUploadMutation($input: UploadBilbyJobMutationInput!) {
                uploadBilbyJob(input: $input) {
                    result {
                        jobId
                    }
                }
            }
        """

        with open(job_archive, 'rb') as f:
            variables = {
                "input": {
                    "uploadToken": self._generate_upload_token(),
                    "details": {
                        "description": description,
                        "private": not public
                    },
                    "jobFile": f
                }
            }

            data = self.request(query=query, variables=variables, authorize=False)

        job_id = data['uploadBilbyJob']['result']['jobId']
        return self.get_job_by_id(job_id)

    def upload_job_directory(self, description, job_directory, public=False):
        """Upload a bilby job to GWCloud by job output directory

        Parameters
        ----------
        description : str
            The description of the job to add to the database

        public : bool
            If the uploaded job should be public or not

        job_directory : str
            The path to the job output directory to upload

        Returns
        -------
        BilbyJob
            The created Bilby job
        """

        # Generate a temporary archive of the job
        path = Path(job_directory)
        with NamedTemporaryFile(dir=job_directory, suffix='.tar.gz') as f:
            with tarfile.open(f.name, "w:gz", compresslevel=2) as tar_handle:
                for item in path.rglob("*"):
                    print(item)
                    tar_handle.add(item, arcname=item.relative_to(path), recursive=False)

            # Upload the archive
            return self.upload_job_archive(description, f.name, public)

    def create_event_id(self, event_id, trigger_id=None, nickname=None, is_ligo_event=False):
        """Create an Event ID that can be assigned to Bilby Jobs

        **INFO**:
        *Event IDs can only be created by a select few users.*

        Parameters
        ----------
        event_id : str
            ID of the event, must be of the form GW123456_123456
        trigger_id : str, optional
            Trigger ID of the event, must be of the form S123456a, by default None
        nickname : str, optional
            Common name used to identify the event, by default None
        is_ligo_event : bool, optional
            Should the event be visible to ligo users only, by default False

        Returns
        -------
        .EventID
            The created Event ID
        """
        query = """
            mutation CreateEventIDMutation($input: EventIDMutationInput!) {
                createEventId (input: $input) {
                    result
                }
            }
        """
        variables = {
            "input": {
                "eventId": event_id,
                "triggerId": trigger_id,
                "nickname": nickname,
                "isLigoEvent": is_ligo_event,
            }
        }
        data = self.request(query=query, variables=variables)
        logger.info(data['createEventId']['result'])
        return self.get_event_id(event_id=event_id)

    def update_event_id(self, event_id, trigger_id=None, nickname=None, is_ligo_event=None):
        """Create an Event ID that can be assigned to Bilby Jobs

        **INFO**:
        *Event IDs can only be updated by a select few users.*

        Parameters
        ----------
        event_id : str
            ID of the event, must be of the form GW123456_123456
        trigger_id : str, optional
            Trigger ID of the event, must be of the form S123456a, by default None
        nickname : str, optional
            Common name used to identify the event, by default None
        is_ligo_event : bool, optional
            Should the event be visible to ligo users only, by default None

        Returns
        -------
        .EventID
            The updated Event ID
        """
        query = """
            mutation UpdateEventIDMutation($input: UpdateEventIDMutationInput!) {
                updateEventId (input: $input) {
                    result
                }
            }
        """
        variables = {
            "input": {
                "eventId": event_id,
                "triggerId": trigger_id,
                "nickname": nickname,
                "isLigoEvent": is_ligo_event,
            }
        }
        data = self.request(query=query, variables=variables)
        logger.info(data['updateEventId']['result'])
        return self.get_event_id(event_id=event_id)

    def delete_event_id(self, event_id):
        """Delete an Event ID

        **INFO**:
        *Event IDs can only be deleted by a select few users.*

        Parameters
        ----------
        event_id : str
            ID of the event, must be of the form GW123456_123456
        """
        query = """
            mutation DeleteEventIDMutation($input: DeleteEventIDMutationInput!) {
                deleteEventId (input: $input) {
                    result
                }
            }
        """
        variables = {
            "input": {
                "eventId": event_id
            }
        }
        data = self.request(query=query, variables=variables)
        logger.info(data['deleteEventId']['result'])

    def get_event_id(self, event_id):
        """Get EventID by the event_id

        Parameters
        ----------
        event_id : str
            Event ID of the form GW123456_123456

        Returns
        -------
        .EventID
            The requested Event ID
        """
        query = """
            query ($eventId: String!){
                eventId (eventId: $eventId) {
                    eventId
                    triggerId
                    nickname
                    isLigoEvent
                }
            }
        """
        if event_id == '':
            return None

        variables = {
            "eventId": event_id
        }
        data = self.request(query=query, variables=variables)
        return EventID(**convert_dict_keys(data['eventId']))

    def get_all_event_ids(self):
        """Obtain a list of all Event IDs

        Parameters
        ----------
        event_id : str
            ID of the event, must be of the form GW123456_123456

        Returns
        -------
        list
            A list of all .EventID objects
        """
        query = """
            query {
                allEventIds {
                    eventId
                    triggerId
                    nickname
                    isLigoEvent
                }
            }
        """
        data = self.request(query=query)
        return [EventID(**convert_dict_keys(event)) for event in data['allEventIds']]



from pathlib import Path
import re


def write_file_at_path(root_path, file_path, file_contents, preserve_directory_structure=True):
    """Write a file at the given path, with the given contents

    Parameters
    ----------
    root_path : str or pathlib.Path
        The base directory
    file_path : pathlib.Path
        The file path within the base directory, including the name of the file
    file_contents : bytes
        The contents of the file
    preserve_directory_structure : bool, optional
        Create any directories present in `file_path`, by default True
    """
    if preserve_directory_structure:
        path = root_path / file_path
    else:
        path = root_path / Path(file_path.name)

    path.parents[0].mkdir(parents=True, exist_ok=True)
    path.write_bytes(file_contents)


def remove_path_anchor(path):
    """Removes the path anchor, making it a relative path

    Parameters
    ----------
    path : pathlib.Path
        Path from which to strip anchor

    Returns
    -------
    Path
        Relative path
    """
    if path.is_absolute():
        return path.relative_to(path.anchor)
    else:
        return path


def rename_dict_keys(input_dict, key_sets):
    """Renames the keys in a dictionary

    Parameters
    ----------
    input_dict : dict
        Dictionary for which to change the keys
    key_sets : list
        list of tuples of the format `(old_key, new_key)`

    Returns
    -------
    dict
        Copy of `input_dict` with old keys subbed for new keys
    """
    output_dict = input_dict.copy()
    for old_key, new_key in key_sets:
        output_dict[new_key] = output_dict.pop(old_key, None)

    return output_dict


def convert_dict_keys(input_dict, key_sets=[]):
    """Convert the keys of a dictionary from camelCase to snake_case

    Parameters
    ----------
    input_dict : dict
        Dictionary for which to convert the keys
    key_sets : list, optional
        List of tuples of the format `(old_key, new_key)` which will also be applied to the dict, by default []

    Returns
    -------
    dict
        Copy of `input_dict` with keys converted from camelCase to snake_case, and optional other key sets exchanged
    """
    convert_key_sets = []
    for old_key in input_dict.keys():
        new_key = re.sub('([A-Z]+)', r'_\1', old_key).lower()
        if old_key != new_key:
            convert_key_sets.append((old_key, new_key))
    return rename_dict_keys(input_dict, convert_key_sets + key_sets)



from dataclasses import dataclass
from enum import Enum


@dataclass
class JobStatus:
    """Contains the status of a job in a more readable format."""
    status: str
    date: str


class TimeRange(Enum):
    """Enum to help with the time range field in the public job search."""
    ANY = "Any time"
    DAY = "Past 24 hours"
    WEEK = "Past week"
    MONTH = "Past month"
    YEAR = "Past year"


class Cluster(Enum):
    """Enum to identify which cluster to submit a job to."""
    DEFAULT = None
    OZSTAR = 'ozstar'
    CIT = 'cit'



from gwcloud_python import BilbyJob, GWCloud, FileReference, FileReferenceList
import pytest
from gwcloud_python.utils import file_filters


@pytest.fixture
def png_data_result():
    return FileReferenceList([
        FileReference(path='data/dir/test1.png', file_size='1', download_token='test_download_token_1'),
        FileReference(path='data/dir/test2.png', file_size='1', download_token='test_download_token_2'),
        FileReference(path='result/dir/test1.png', file_size='1', download_token='test_download_token_3'),
        FileReference(path='result/dir/test2.png', file_size='1', download_token='test_download_token_4'),
    ])


@pytest.fixture
def png_extra():
    return FileReferenceList([
        FileReference(path='test1.png', file_size='1', download_token='test_download_token_5'),
        FileReference(path='test2.png', file_size='1', download_token='test_download_token_6'),
        FileReference(path='arbitrary/dir/test1.png', file_size='1', download_token='test_download_token_7'),
        FileReference(path='arbitrary/dir/test2.png', file_size='1', download_token='test_download_token_8'),
    ])


@pytest.fixture
def corner():
    return FileReferenceList([
        FileReference(path='test1_corner.png', file_size='1', download_token='test_download_token_9'),
        FileReference(path='test2_corner.png', file_size='1', download_token='test_download_token_10'),
    ])


@pytest.fixture
def config():
    return FileReferenceList([
        FileReference(path='a_config_complete.ini', file_size='1', download_token='test_download_token_11'),
    ])


@pytest.fixture
def json():
    return FileReferenceList([
        FileReference(path='result/dir/a_merge_result.json', file_size='1', download_token='test_download_token_12'),
    ])


@pytest.fixture
def index():
    return FileReferenceList([
        FileReference(path='index.html', file_size='1', download_token='test_download_token_13'),
    ])


@pytest.fixture
def png(png_data_result, png_extra, corner):
    return png_data_result + png_extra + corner


@pytest.fixture
def default(png_data_result, config, json, index):
    return png_data_result + config + json + index


@pytest.fixture
def full(png, config, json, index):
    return png + config + json + index


@pytest.fixture
def setup_mock_gwcloud(mocker, full):
    def mock_init(self, token, endpoint='test.endpoint.com'):
        pass

    def mock_get_files_by_job_id(self, job_id):
        return full, {'bilbyResultFiles': {'isUploadedJob': False}}

    mocker.patch('gwcloud_python.gwcloud.GWCloud.__init__', mock_init)
    mocker.patch('gwcloud_python.gwcloud.GWCloud._get_files_by_job_id', mock_get_files_by_job_id)

    return GWCloud(token='test_token')


@pytest.fixture
def bilby_job(setup_mock_gwcloud):
    return BilbyJob(
        client=setup_mock_gwcloud,
        job_id='test_id',
        name='TestName',
        description='Test description',
        job_status={
            'name': 'Completed',
            'date': '2021-12-02'
        }
    )


def test_bilby_job_full_file_list(bilby_job, full):
    assert bilby_job.get_full_file_list() == full


def test_bilby_job_file_filters(bilby_job, default, png, corner, config):
    assert file_filters.sort_file_list(bilby_job.get_default_file_list()) == file_filters.sort_file_list(default)
    assert file_filters.sort_file_list(bilby_job.get_png_file_list()) == file_filters.sort_file_list(png)
    assert file_filters.sort_file_list(bilby_job.get_corner_plot_file_list()) == file_filters.sort_file_list(corner)
    assert file_filters.sort_file_list(bilby_job.get_config_file_list()) == file_filters.sort_file_list(config)


def test_register_file_list_filter(bilby_job, index):
    def get_html_file(file_list):
        return [f for f in file_list if f.path.suffix == '.html']

    assert getattr(bilby_job, 'get_index_file_list', None) is None
    assert getattr(bilby_job, 'get_index_files', None) is None
    assert getattr(bilby_job, 'save_index_files', None) is None

    BilbyJob.register_file_list_filter('index', get_html_file)

    assert getattr(bilby_job, 'get_index_file_list', None) is not None
    assert getattr(bilby_job, 'get_index_files', None) is not None
    assert getattr(bilby_job, 'save_index_files', None) is not None

    assert bilby_job.get_index_file_list() == index



from dataclasses import dataclass, field
from collections import UserList
from pathlib import Path
from .utils import remove_path_anchor, file_filters


@dataclass
class FileReference:
    """Object used to facilitate simpler downloading of files.
    """
    path: str
    file_size: int = field(repr=False)
    download_token: str = field(repr=False)

    def __post_init__(self):
        self.path = remove_path_anchor(Path(self.path))
        self.file_size = int(self.file_size)


class FileReferenceList(UserList):
    """Used to store FileReference objects and provide simple methods with which to obtain their data.
    As a subclass of ~collections.UserList, this class contains the same functionality as a regular list.
    It also contains several other useful methods.

    Parameters
    ----------
    initlist : list
        List of FileReference objects
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for item in self.data:
            if not isinstance(item, FileReference):
                raise Exception('FileReferenceList must be initialised with a list of FileReference objects')

    def append(self, item):
        """Use as you would list.append()

        Parameters
        ----------
        item : FileReference
            If this is not a FileReference instance, the method will fail

        Raises
        ------
        Exception
            Raised if item is not a FileReference
        """
        if not isinstance(item, FileReference):
            raise Exception('Appended item must be a FileReference object')
        self.data.append(item)

    def filter_list(self, file_filter_fn, *args, **kwargs):
        """Create a subset of this list by filtering the contents with a function.

        Parameters
        ----------
        file_filter_fn : function
            Must take a list of FileReference objects and return only those that are desired

        Returns
        -------
        FileReferenceList
            Filtered list
        """
        return FileReferenceList(file_filter_fn(self.data, *args, **kwargs))

    def filter_list_by_path(self, directory=None, name=None, extension=None):
        """Create a subset of this list by filtering the contents based on their path attributes

        Parameters
        ----------
        directory : str, optional
            Matches any of the directories in the file path, by default None
        name : str, optional
            Matches the name of the file, by default None
        extension : str, optional
            Matches the file extension, by default None

        Returns
        -------
        FileReferenceList
            Filtered list
        """
        return self.filter_list(file_filters.custom_path_filter, directory, name, extension)

    def get_total_bytes(self):
        """Sum the total size of each file represented in the list

        Returns
        -------
        int
            Total size of all files
        """
        total_bytes = 0
        for ref in self.data:
            total_bytes += ref.file_size

        return total_bytes

    def get_tokens(self):
        """Get all the download tokens in a list

        Returns
        -------
        list
            List of download tokens
        """
        return [ref.download_token for ref in self.data]

    def get_paths(self):
        """Get all the file paths in a list

        Returns
        -------
        list
            List of file paths
        """
        return [ref.path for ref in self.data]

    def get_output_paths(self, root_path, preserve_directory_structure=True):
        """Get all the file paths modified to give them a base directory.
        Can also optionally remove any existing directory structure

        Parameters
        ----------
        root_path : str or ~pathlib.Path
            Directory to add to the beginning of the file paths
        preserve_directory_structure : bool, optional
            Retain existing directory structure in the file paths, by default True

        Returns
        -------
        list
            List of output file paths
        """
        paths = []
        for ref in self.data:
            if preserve_directory_structure:
                paths.append(root_path / ref.path)
            else:
                paths.append(root_path / Path(ref.path.name))
        return paths



from dataclasses import dataclass


@dataclass
class EventID:
    """Object used to help with abstraction of Event IDs. Currently a glorified dictionary."""
    event_id: str
    trigger_id: str = None
    nickname: str = None
    is_ligo_event: bool = False



import pytest
from pathlib import Path
from gwcloud_python.utils import identifiers


@pytest.fixture
def setup_paths():
    return {
        'png': Path('test.png'),

        'html': Path('test.html'),
        'dir_html': Path('this/is/a/test.html'),

        'no_suffix': Path('this/is/a/test'),
        'png_dir': Path('this/is/not/a/png'),
        'html_dir': Path('this/is/not/a/html'),

        'corner': Path('this/is/a_corner.png'),
        'corner_bad_ext': Path('this/is/a_corner.txt'),
        'corner_bad_pattern': Path('this/is/a_coroner.png'),

        'config': Path('this/is/a_config_complete.ini'),
        'config_bad_ext': Path('this/is/a_config_complete.txt'),
        'config_bad_pattern': Path('this/is/a_config_incomplete.ini'),

        'merged_json': Path('this/is/a_merge_result.json'),
        'merged_json_bad_ext': Path('this/is/a_merge_result.txt'),
        'merged_json_bad_pattern': Path('this/is/a_merged_result.txt'),

        'unmerged_json': Path('this/is/a_result.json'),
        'unmerged_json_bad_ext': Path('this/is/a_result.txt'),
        'unmerged_json_bad_pattern': Path('this/is/a_resalt.json'),

        'absolute_data_path': Path('/data/this/is/a/test'),
        'data_path': Path('data/this/is/a/test'),
        'data_dir': Path('data/'),

        'absolute_result_path': Path('/result/this/is/a/test'),
        'result_path': Path('result/this/is/a/test'),
        'result_dir': Path('result/'),

        'data_dir_png': Path('data/dir/test.png'),
        'data_png': Path('data.png'),

        'result_dir_png': Path('result/dir/test.png'),
        'result_png': Path('result.png'),

        'result_merged_json': Path('result/dir/a_merge_result.json'),
        'result_unmerged_json': Path('result/dir/a_result.json'),
    }


@pytest.fixture
def setup_identifiers():
    return [
        (
            identifiers.html_file,
            ['html', 'dir_html']
        ),
        (
            identifiers.png_file,
            ['png', 'corner', 'corner_bad_pattern', 'data_dir_png', 'data_png', 'result_dir_png', 'result_png']
        ),
        (
            identifiers.data_dir,
            ['data_path', 'data_dir', 'data_dir_png']
        ),
        (
            identifiers.result_dir,
            ['result_path', 'result_dir', 'result_dir_png', 'result_merged_json', 'result_unmerged_json']
        ),
        (
            identifiers.config_file,
            ['config']
        ),
        (
            identifiers.merged_json_file,
            ['merged_json', 'result_merged_json']
        ),
        (
            identifiers.unmerged_json_file,
            ['unmerged_json', 'result_unmerged_json', 'merged_json', 'result_merged_json']
        ),
        (
            identifiers.corner_plot_file,
            ['corner']
        ),
        (
            identifiers.data_png_file,
            ['data_dir_png']
        ),
        (
            identifiers.result_png_file,
            ['result_dir_png']
        ),
        (
            identifiers.result_merged_json_file,
            ['result_merged_json']
        ),
        (
            identifiers.result_json_file,
            ['result_merged_json', 'result_unmerged_json']
        ),
    ]


@pytest.fixture
def check_identifier(setup_paths):
    def _check_identifier(identifier, true_path_keys):
        true_paths = [value for key, value in setup_paths.items() if key in true_path_keys]
        false_paths = [value for key, value in setup_paths.items() if key not in true_path_keys]

        for path in true_paths:
            assert identifier(path) is True

        for path in false_paths:
            assert identifier(path) is False

    return _check_identifier


def test_identifiers(setup_identifiers, check_identifier):
    for identifier, true_path_keys in setup_identifiers:
        check_identifier(identifier, true_path_keys)



from gwcloud_python import FileReference, FileReferenceList
import pytest
from pathlib import Path
from gwcloud_python.utils import remove_path_anchor


@pytest.fixture
def setup_dicts():
    return [
        {'path': 'data/dir/test1.png', 'file_size': '1', 'download_token': 'test_download_token_1'},
        {'path': 'data/dir/test2.png', 'file_size': '1', 'download_token': 'test_download_token_2'},
        {'path': 'result/dir/test1.png', 'file_size': '1', 'download_token': 'test_download_token_3'},
        {'path': 'result/dir/test2.png', 'file_size': '1', 'download_token': 'test_download_token_4'},
    ]


def test_file_reference(setup_dicts):
    for file_dict in setup_dicts:
        ref = FileReference(**file_dict)
        assert ref.path == remove_path_anchor(Path(file_dict['path']))
        assert ref.file_size == int(file_dict['file_size'])
        assert ref.download_token == file_dict['download_token']


def test_file_reference_list(setup_dicts):
    file_references = [FileReference(**file_dict) for file_dict in setup_dicts]
    file_reference_list = FileReferenceList(file_references)

    for i, ref in enumerate(file_reference_list):
        assert ref.path == file_references[i].path
        assert ref.file_size == file_references[i].file_size
        assert ref.download_token == file_references[i].download_token

    assert file_reference_list.get_total_bytes() == sum([ref.file_size for ref in file_references])
    assert file_reference_list.get_tokens() == [ref.download_token for ref in file_references]
    assert file_reference_list.get_paths() == [ref.path for ref in file_references]



from gwcloud_python import GWCloud, FileReference, JobStatus
from gwcloud_python.utils import convert_dict_keys
import pytest


@pytest.fixture
def setup_mock_gwdc(mocker):
    def mock_gwdc(request_data):
        def mock_init(self, token, endpoint):
            pass

        def mock_request(self, query, variables=None, headers=None):
            return request_data

        mocker.patch('gwdc_python.gwdc.GWDC.__init__', mock_init)
        mocker.patch('gwdc_python.gwdc.GWDC.request', mock_request)

    return mock_gwdc


@pytest.fixture
def single_job_request(setup_mock_gwdc):
    job_data = {
        "id": 1,
        "name": "test_name",
        "description": "test description",
        "userId": 1,
        "jobStatus": {
            "name": "Completed",
            "date": "2021-12-02"
        }
    }
    setup_mock_gwdc({"bilbyJob": job_data})
    return job_data


@pytest.fixture
def multi_job_request(setup_mock_gwdc):
    def modify_query_name(query_name):
        job_data_1 = {
            "id": 1,
            "name": "test_name_1",
            "description": "test description 1",
            "userId": 1,
            "jobStatus": {
                "name": "Completed",
                "date": "2021-01-01"
            }
        }

        job_data_2 = {
            "id": 2,
            "name": "test_name_2",
            "description": "test description 2",
            "userId": 2,
            "jobStatus": {
                "name": "Completed",
                "date": "2021-02-02"
            }
        }

        job_data_3 = {
            "id": 3,
            "name": "test_name_3",
            "description": "test description 3",
            "userId": 3,
            "jobStatus": {
                "name": "Error",
                "date": "2021-03-03"
            }
        }

        setup_mock_gwdc({
            query_name: {
                "edges": [
                    {"node": job_data_1},
                    {"node": job_data_2},
                    {"node": job_data_3},
                ]
            }
        })

        return [job_data_1, job_data_2, job_data_3]

    return modify_query_name


@pytest.fixture
def job_file_request(setup_mock_gwdc):
    job_file_data_1 = {
        "path": "path/to/test.png",
        "fileSize": "1",
        "downloadToken": "test_token_1",
        "isDir": False
    }

    job_file_data_2 = {
        "path": "path/to/test.json",
        "fileSize": "10",
        "downloadToken": "test_token_2",
        "isDir": False
    }

    job_file_data_3 = {
        "path": "path/to/test",
        "fileSize": "100",
        "downloadToken": "test_token_3",
        "isDir": True
    }

    setup_mock_gwdc({
        "bilbyResultFiles": {
            "files": [
                job_file_data_1,
                job_file_data_2,
                job_file_data_3
            ],
            "isUploadedJob": False
        }
    })

    return [job_file_data_1, job_file_data_2, job_file_data_3]


@pytest.fixture
def user_jobs(multi_job_request):
    return multi_job_request('bilbyJobs')


def test_get_job_by_id(single_job_request):
    gwc = GWCloud(token='my_token')

    job = gwc.get_job_by_id('job_id')

    assert job.job_id == single_job_request["id"]
    assert job.name == single_job_request["name"]
    assert job.description == single_job_request["description"]
    assert job.status == JobStatus(
        status=single_job_request["jobStatus"]["name"],
        date=single_job_request["jobStatus"]["date"]
    )
    assert job.other['user_id'] == single_job_request["userId"]


def test_get_user_jobs(user_jobs):
    gwc = GWCloud(token='my_token')

    jobs = gwc.get_user_jobs()

    assert jobs[0].job_id == user_jobs[0]["id"]
    assert jobs[0].name == user_jobs[0]["name"]
    assert jobs[0].description == user_jobs[0]["description"]
    assert jobs[0].status == JobStatus(
        status=user_jobs[0]["jobStatus"]["name"],
        date=user_jobs[0]["jobStatus"]["date"]
    )
    assert jobs[0].other['user_id'] == user_jobs[0]["userId"]

    assert jobs[1].job_id == user_jobs[1]["id"]
    assert jobs[1].name == user_jobs[1]["name"]
    assert jobs[1].description == user_jobs[1]["description"]
    assert jobs[1].status == JobStatus(
        status=user_jobs[1]["jobStatus"]["name"],
        date=user_jobs[1]["jobStatus"]["date"]
    )
    assert jobs[1].other['user_id'] == user_jobs[1]["userId"]

    assert jobs[2].job_id == user_jobs[2]["id"]
    assert jobs[2].name == user_jobs[2]["name"]
    assert jobs[2].description == user_jobs[2]["description"]
    assert jobs[2].status == JobStatus(
        status=user_jobs[2]["jobStatus"]["name"],
        date=user_jobs[2]["jobStatus"]["date"]
    )
    assert jobs[2].other['user_id'] == user_jobs[2]["userId"]


def test_gwcloud_files_by_job_id(job_file_request):
    gwc = GWCloud(token='my_token')

    file_list, is_uploaded_job = gwc._get_files_by_job_id('arbitrary_job_id')

    for i, ref in enumerate(file_list):
        assert ref == FileReference(**convert_dict_keys(job_file_request[i]))



import logging
from .utils import file_filters
from .helpers import JobStatus
from .event_id import EventID

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
logger.addHandler(ch)


class BilbyJob:
    """
    BilbyJob class is useful for interacting with the Bilby jobs returned from a call to the GWCloud API.
    It is primarily used to store job information and obtain files related to the job.

    Parameters
    ----------
    client : ~gwcloud_python.gwcloud.GWCloud
        A reference to the GWCloud object instance from which the BilbyJob was created
    job_id : str
        The id of the Bilby job, required to obtain the files associated with it
    name : str
        Job name
    description : str
        Job description
    job_status : dict
        Status of job, should have 'name' and 'date' keys corresponding to the status code and when it was produced
    kwargs : dict, optional
        Extra arguments, stored in `other` attribute
    """

    DEFAULT_FILE_LIST_FILTERS = {
        'default': file_filters.default_filter,
        'config': file_filters.config_filter,
        'png': file_filters.png_filter,
        'corner_plot': file_filters.corner_plot_filter,
    }

    def __init__(self, client, job_id, name, description, job_status, **kwargs):
        self.client = client
        self.job_id = job_id
        self.name = name
        self.description = description
        self.status = JobStatus(status=job_status['name'], date=job_status['date'])
        self.other = kwargs
        self.is_uploaded_job = None

    def __repr__(self):
        return f"BilbyJob(name={self.name}, job_id={self.job_id})"

    def get_full_file_list(self):
        """Get information for all files associated with this job

        Returns
        -------
        .FileReferenceList
            Contains FileReference instances for each of the files associated with this job
        """
        result, self.is_uploaded_job = self.client._get_files_by_job_id(self.job_id)
        return result

    def get_files_by_reference(self, file_references):
        """Obtain the data for a FileReferenceList

        Parameters
        ----------
        file_references : .FileReferenceList
            Contains FileReference instances for each of the files to be downloaded

        Returns
        -------
        list
            List of tuples with the file path and downloaded file contents as a byte string
        """
        return self.client._get_files_by_reference(self.job_id, file_references, self.is_uploaded_job)

    def save_files_by_reference(self, file_references, root_path, preserve_directory_structure=True):
        """Save the files represented in a FileReferenceList

        Parameters
        ----------
        file_references : .FileReferenceList
            Contains FileReference instances for each of the files to be downloaded and saved
        root_path : str or ~pathlib.Path
            Directory into which to save the files
        preserve_directory_structure : bool, optional
            Whether or not the files should retain the directory structure in which they are downloaded, by default True

        Returns
        -------
        str
            Success message
        """
        return self.client._save_files_by_reference(
            self.job_id, file_references, root_path, preserve_directory_structure, self.is_uploaded_job
        )

    @classmethod
    def register_file_list_filter(cls, name, file_list_filter_fn):
        """Register a function used to filter the file list.
        This will create three methods on the class using this filter function:

        - get_{name}_file_list
        - get_{name}_files
        - save_{name}_files

        where {name} is the input name string.

        Parameters
        ----------
        name : str
            String used to name the added methods
        file_list_filter_fn : function
            A function that takes in the full file list and returns only the desired entries from the list
        """
        _register_file_list_filter(name, file_list_filter_fn)
        cls.DEFAULT_FILE_LIST_FILTERS[f'{name}'] = file_list_filter_fn

    def set_event_id(self, event_id=None):
        """Set the Event ID of a Bilby Job

        Parameters
        ----------
        event_id : EventID or str
            The desired Event ID, by default None
        """
        query = """
            mutation BilbyJobEventIDMutation($input: UpdateBilbyJobMutationInput!) {
                updateBilbyJob(input: $input) {
                    result
                }
            }
        """

        if isinstance(event_id, EventID):
            new_event_id = event_id.event_id
        elif isinstance(event_id, str):
            new_event_id = event_id
        elif event_id is None:
            new_event_id = ''
        else:
            raise Exception('Parameter event_id must be an EventID, a string or None')

        variables = {
            "input": {
                "jobId": self.job_id,
                "eventId": new_event_id
            }
        }

        data = self.client.request(query=query, variables=variables)
        self.other['event_id'] = self.client.get_event_id(event_id=new_event_id)
        logger.info(data['updateBilbyJob']['result'])


def _register_file_list_filter(name, file_list_filter_fn):
    spaced_name = name.replace('_', ' ')

    def _get_file_list_subset(self):
        full_list = self.get_full_file_list()
        return full_list.filter_list(file_list_filter_fn)

    file_list_fn_name = f'get_{name}_file_list'
    file_list_fn = _get_file_list_subset
    file_list_fn.__doc__ = f"""Get information for the {spaced_name} files associated with this job

        Returns
        -------
        .FileReferenceList
            Contains FileReference instances holding information on the {spaced_name} files
    """
    setattr(BilbyJob, file_list_fn_name, file_list_fn)

    def _get_files(self):
        file_list = _get_file_list_subset(self)
        return self.get_files_by_reference(file_list)

    files_fn_name = f'get_{name}_files'
    files_fn = _get_files
    files_fn.__doc__ = f"""Download the content of all the {spaced_name} files.

        **WARNING**:
        *As the file contents are stored in memory, we suggest being cautious about the size of files being downloaded.
        If the files are large or very numerous, it is suggested to save the files and read them as needed instead.*

        Returns
        -------
        list
            List containing tuples of the file path and associated file contents
    """
    setattr(BilbyJob, files_fn_name, files_fn)

    def _save_files(self, root_path, preserve_directory_structure=True):
        file_list = _get_file_list_subset(self)
        return self.save_files_by_reference(file_list, root_path, preserve_directory_structure)

    save_fn_name = f'save_{name}_files'
    save_fn = _save_files
    save_fn.__doc__ = f"""Download and save the {spaced_name} files.

        Parameters
        ----------
        root_path : str or ~pathlib.Path
            The base directory into which the files will be saved
        preserve_directory_structure : bool, optional
            Save the files in the same structure that they were downloaded in, by default True
    """
    setattr(BilbyJob, save_fn_name, save_fn)


for name, file_filter in BilbyJob.DEFAULT_FILE_LIST_FILTERS.items():
    _register_file_list_filter(name, file_filter)



def _file_suffix(file_path, suffix):
    return file_path.suffix == f'.{suffix}'


def _file_base_dir(file_path, directory):
    return file_path.parts[0] == directory


def _file_dir(file_path, directory):
    return directory in file_path.parts[:-1]


def _file_name(file_path, name):
    return name in file_path.name


def _file_end(file_path, end):
    return file_path.name.endswith(end)


def html_file(file_path):
    """Checks to see if the given file path points to a HTML file

    Parameters
    ----------
    file_path : ~pathlib.Path
        File path to check

    Returns
    -------
    bool
        True if path points to a HTML file, False otherwise
    """
    return _file_suffix(file_path, 'html')


def png_file(file_path):
    """Checks to see if the given file path ends points to a PNG file

    Parameters
    ----------
    file_path : ~pathlib.Path
        File path to check

    Returns
    -------
    bool
        True if path points to a PNG file, False otherwise
    """
    return _file_suffix(file_path, 'png')


def data_dir(file_path):
    """Checks to see if the given file path starts with 'data' directory

    Parameters
    ----------
    file_path : ~pathlib.Path
        File path to check

    Returns
    -------
    bool
        True if path starts with 'data' directory, False otherwise
    """
    return _file_base_dir(file_path, 'data')


def result_dir(file_path):
    """Checks to see if the given file path starts with 'result' directory

    Parameters
    ----------
    file_path : ~pathlib.Path
        File path to check

    Returns
    -------
    bool
        True if path starts with 'result' directory, False otherwise
    """
    return _file_base_dir(file_path, 'result')


def config_file(file_path):
    """Checks to see if the given file path points towards the config file

    Parameters
    ----------
    file_path : ~pathlib.Path
        File path to check

    Returns
    -------
    bool
        True if path points to config file, False otherwise
    """
    return _file_end(file_path, '_config_complete.ini')


def merged_json_file(file_path):
    """Checks to see if the given file path points towards a merged JSON file

    Parameters
    ----------
    file_path : ~pathlib.Path
        File path to check

    Returns
    -------
    bool
        True if path points to merged JSON file, False otherwise
    """
    return _file_end(file_path, '_merge_result.json')


def unmerged_json_file(file_path):
    """Checks to see if the given file path points towards the JSON file if not run in parallel

    Parameters
    ----------
    file_path : ~pathlib.Path
        File path to check

    Returns
    -------
    bool
        True if path points to JSON file, False otherwise
    """
    return _file_end(file_path, '_result.json')


def corner_plot_file(file_path):
    """Checks to see if the given file path points towards a corner plot file

    Parameters
    ----------
    file_path : ~pathlib.Path
        File path to check

    Returns
    -------
    bool
        True if path points to corner plot file, False otherwise
    """
    return _file_end(file_path, '_corner.png')


def data_png_file(file_path):
    """Checks to see if the given file path points to a PNG file in the 'data' directory

    Parameters
    ----------
    file_path : ~pathlib.Path
        File path to check

    Returns
    -------
    bool
        True if path points to PNG file in 'data' directory, False otherwise
    """
    return (data_dir(file_path) and png_file(file_path))


def result_png_file(file_path):
    """Checks to see if the given file path points to a PNG file in the 'result' directory

    Parameters
    ----------
    file_path : ~pathlib.Path
        File path to check

    Returns
    -------
    bool
        True if path points to PNG file in 'result' directory, False otherwise
    """
    return (result_dir(file_path) and png_file(file_path))


def result_merged_json_file(file_path):
    """Checks to see if the given file path points to a merged JSON file in the 'result' directory.

    Parameters
    ----------
    file_path : ~pathlib.Path
        File path to check

    Returns
    -------
    bool
        True if path points to the desired JSON file in 'result' directory, False otherwise
    """
    return (result_dir(file_path) and merged_json_file(file_path))


def result_json_file(file_path):
    """Checks to see if the given file path points to a result JSON file in the 'result' directory.

    Parameters
    ----------
    file_path : ~pathlib.Path
        File path to check

    Returns
    -------
    bool
        True if path points to the desired JSON file in 'result' directory, False otherwise
    """
    return (result_dir(file_path) and unmerged_json_file(file_path))



from .gwcloud import GWCloud
from .bilby_job import BilbyJob
from .event_id import EventID
from .file_reference import FileReference, FileReferenceList
from .helpers import TimeRange, Cluster, JobStatus

try:
    from importlib.metadata import version
except ModuleNotFoundError:
    from importlib_metadata import version
__version__ = version('gwcloud_python')


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gwcloud_python',
 'gwcloud_python.tests',
 'gwcloud_python.utils',
 'gwcloud_python.utils.tests']

package_data = \
{'': ['*']}

install_requires = \
['graphene-file-upload>=1.3.0,<2.0.0',
 'gwdc-python>=0.3.0,<0.4.0',
 'importlib-metadata>=4.5.0,<5.0.0',
 'jwt>=1.2.0,<2.0.0',
 'requests>=2.25.1,<3.0.0',
 'tqdm>=4.61.2,<5.0.0']

extras_require = \
{'docs': ['Sphinx>=4.0.2,<5.0.0', 'sphinx-rtd-theme>=0.5.2,<0.6.0']}

setup_kwargs = {
    'name': 'gwcloud-python',
    'version': '1.6.1',
    'description': 'Wrapper of GWDC API, used for interacting with the GWCloud endpoints',
    'long_description': "GWCloud Python API\n==================\n\n`GWCloud <https://gwcloud.org.au/>`_ is a service used to handle both the submission of `Bilby <https://pypi.org/project/bilby/>`_ jobs to a supercomputer queue and the obtaining of the results produced by these jobs.\nWhile there is a web interface for this service, which is recommended for beginners, this package can be used to allow Bilby job submission and manipulation from Python scripts.\n\nCheck out the `documentation <https://gwcloud-python.readthedocs.io/en/latest/>`_ for more information.\n\nInstallation\n------------\n\nThe gwcloud-python package can be installed with\n\n::\n\n    pip install gwcloud-python\n\n\nExample\n-------\n\n::\n\n    >>> from gwcloud_python import GWCloud\n    >>> gwc = GWCloud(token='<user_api_token_here>')\n    >>> job = gwc.get_preferred_job_list()[0]\n    >>> job.save_corner_plot_files()\n\n    100%|██████████████████████████████████████| 3.76M/3.76M [00:00<00:00, 5.20MB/s]\n    All 2 files saved!\n",
    'author': 'Thomas Reichardt',
    'author_email': 'treichardt@swin.edu.au',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/gravitationalwavedc/gwcloud_python',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
